#
#  Text substitution module.
#
#  This file is part of the SuSE help system.
#
#  Copyright (C) 2002  SuSE Linux AG, Nuernberg
#
#  Author: Cornelius Schumacher <cschum@suse.de>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package substitution;
require Exporter;

our @ISA = qw( Exporter );
our @EXPORT = qw( substituteSymbols substituteSymbolsFromFile readSymbols
                  registerSymbolFunction registerSymbol );
our @VERSION = 1.00;

#$dbg = 1;

=head1 Substitution module

=over

=item substituteSymbolsFromFile($)

Perform symbol substition on the text in the supplied filename.  Returns the substituted text.

=cut

sub substituteSymbolsFromFile($)
{
  my $filename = shift;

  if ( !open( IN, "<$filename" ) ) {
    print STDERR "Error: Can't open file '$filename'.\n";
    exit 1;
  }

  $data = join "", <IN>;

  close IN;

  return substituteSymbols( $data );
}

sub substituteSymbols($);

=item substituteSymbols($)

Partial document substitution, expands --SYMBOL:ARGS-- occurrences.
This function recursively expands symbols in the format --SYMBOL:ARGS in the input string, using a symbols map initialised with readSymbols() first.  If substituteSymbols() is called without initialising the symbols map, the program will exit with an error message.  The function has simple infinite recursion checking but will not detect recursive calls where A expands to B and B expands to A.

=cut

sub substituteSymbols($)
{
  if ( !$symbolsRead ) {
    print STDERR "Error. You have to call Substituion::readSymbols() before " .
                 " Substitution::substituteSymbols().\n";
    exit 1;
  }

  my $in = shift;

#  print "-----SUBSTITUTE--start--\n$in\n\-----SUBSTITUTE---end---\n";

  my $out;
  my $start = 0;

  local $_ = $in;
  while ( /--(\w+:.*?)--/sgc ) {
    my $expr = $1;

    if ( $lastcall[ $level - 1 ] && $lastcall[ $level - 1 ] eq $expr ) {
      print STDERR "Warning. Infinite recursion in symbol '$expr'\n";
      return "--$expr--";
      exit 1;
    }
    $lastcall[ ++$level ] = $expr;

    $out .= substr( $in, $start, pos() - $start - length( $expr ) - 4 );
    $start = pos();

    $out .= substituteSymbols( parseSymbol( $expr ) );

    $lastcall[ $level-- ] = "";

    pos = $start;
  }
  $out .= substr( $in, pos );

#  print "^------\n";

  return $out;
}

=item parseSymbol($)

Return value of given symbol.
This looks up the supplied symbol in the simple symbol map and the function 
map, checks that the symbol has the correct form, and returns the result.
If the symbol is not found or is misformed, it returns null.
If the symbol expands to a function, the function is called with the supplied argument and we return the result.

=cut

sub parseSymbol()
{
  my $expr = shift;

  if ( $expr !~ /(.*?):(.*)/ ) { return; }
  my $symbol = $1;
  my $arg = $2;

  $dbg && print STDERR "FOUND symbol: '$symbol'  Arg: '$arg'\n";

  if ( exists $symbolMap{ $symbol } ) {
    return $symbolMap{ $symbol };
  }
  if ( exists $symbolFunctionMap{ $symbol } ) {
    $dbg && print STDERR "FUNCTION call\n";
    $function = $symbolFunctionMap{ $symbol };
    return &$function( $arg );
  }
  return "";
}

=item readSymbols($)

Read symbol map from given file and put data into global hash symbolMap.

=cut

sub readSymbols($)
{
  my $symbolFilename = shift;

  if ( !open( SYMBOLS, "<$symbolFilename" ) ) {
    print STDERR "Error: Can't open symbol file '$symbolFilename'\n";
    exit 1;
  }

  my $state = 0;  # 0: Look for symbol name, 1: get one-line value
                  # 2: get multi line value
  my ( $name, $value );
  while( <SYMBOLS> ) {
    if ( $state == 0 ) {
      if ( /^\s*$/ ) { next; }
      if ( /<<\s+EOT/ ) { $state = 2; }
      else { $state = 1; }
      /(\w+)\s*/;
      $name = $1;
      $dbg && print STDERR "DEFINE symbol name: $name\n";
      next;
    }
    if ( $state == 1 ) {
      chop;
      $symbolMap{ $name } = $_;
      $state = 0;
      $name = "";
      $value = "";
      next;
    }
    if ( $state == 2 ) {
      if ( /^EOT/ ) {
        $symbolMap{ $name } = $value;
        $state = 0;
        $name = "";
        $value = "";
      } else {
        $value .= $_;
      }
      next;
    }
    print STDERR "Illegal state $state. Aborting.\n";
    exit 1;
  }

  $symbolsRead = 1;
}

=item registerSymbol($$)

add the given symbol, value pair to the symbol map

=cut

sub registerSymbol($$)
{
  my $name = shift;
  my $value = shift;
  
  $symbolMap{ $name } = $value;
}

=item registerSymbolFunction($$)

add the given symbol, function pair to the function map

=cut

sub registerSymbolFunction($$)
{
  my $name = shift;
  my $function = shift;

  $symbolFunctionMap{ $name } = $function;
}
